
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;
import static java.util.concurrent.Executors.newFixedThreadPool;

public class ProfilingMergeSort {

// test results on laptop i7 processor (two cores with hyperthreading)    
//  Size 20000000 modulo 1 with time: 10191
//  Size 20000000 modulo 2 with time: 5721
//  Size 20000000 modulo 3 with time: 4375
//  Size 20000000 modulo 4 with time: 4354
//  Size 20000000 modulo 5 with time: 4191
//  Size 20000000 modulo 6 with time: 4078
//  Size 20000000 modulo 7 with time: 4229
//  Size 20000000 modulo 8 with time: 4284
//  Size 20000000 modulo 9 with time: 3928
//  Size 20000000 modulo 10 with time: 3982
//  Size 20000000 modulo 20 with time: 3827
//  Size 20000000 modulo 50 with time: 3765
    
    private int sizeToPar;

    private ExecutorService threadPool;

    public static void main(String[] args) {

        int size = 20000000;
        ProfilingMergeSort sort = new ProfilingMergeSort();

        List<Integer> list = new ArrayList<>(size);
        Random random = new Random();

        for (int i = 0; i < size; i++) {
            list.add(random.nextInt());
        }

        sort.test(list, 1);
        sort.test(list, 2);
        sort.test(list, 3);
        sort.test(list, 4);
// better CPU?
        sort.test(list, 5);             
        sort.test(list, 6);             
        sort.test(list, 7);             
        sort.test(list, 8);             
        sort.test(list, 9);             
        sort.test(list, 10);
        sort.test(list, 20);
        sort.test(list, 50);
    }

    public void test(List<Integer> list, int divider) {

        sizeToPar = list.size() / divider;

        threadPool = newFixedThreadPool(list.size());

        long time = System.currentTimeMillis();
        mergeSort(list);
        System.out.println("Size " + list.size() + " modulo " + divider + " with time: " + (System.currentTimeMillis() - time));

        threadPool.shutdown(); // shutdown the pool.

    }

    public List<Integer> mergeSort(List<Integer> list) {
        //If list is empty; no need to do anything
        if (list.size() <= 1) {
            return list;
        }

        //Split the array in half in two parts
        List<Integer> first = list.subList(0, list.size() / 2);
        List<Integer> second = list.subList(first.size(), list.size());

        if (list.size() > sizeToPar) {
            try {
                Future<List<Integer>> firstR = threadPool.submit(() -> mergeSort(first));
                Future<List<Integer>> secondR = threadPool.submit(() -> mergeSort(second));
                return merge(firstR.get(), secondR.get());
            } catch (InterruptedException | ExecutionException ex) {
                Logger.getLogger(ProfilingMergeSort.class.getName()).log(Level.SEVERE, null, ex);
            }
        }

        //Sort each half recursively
        List<Integer> mergeSortFirst = mergeSort(first);
        List<Integer> mergeSortSecond = mergeSort(second);

        //Merge both halves together
        return merge(mergeSortFirst, mergeSortSecond);
    }

    private List<Integer> merge(List<Integer> first, List<Integer> second) {

        //Index Position in first array - starting with first element
        int iFirst = 0;

        //Index Position in second array - starting with first element
        int iSecond = 0;

        List<Integer> result = new ArrayList<>(first.size() + second.size());

        //Compare elements at iFirst and iSecond, 
        //and move smaller element at iMerged
        while (iFirst < first.size() && iSecond < second.size()) {
            if (first.get(iFirst).compareTo(second.get(iSecond)) < 0) {
                result.add(first.get(iFirst));
                iFirst++;
            } else {
                result.add(second.get(iSecond));
                iSecond++;
            }
        }
        //copy remaining elements from both halves - each half will have already sorted elements
        for (int i = iFirst; i < first.size(); i++) {
            result.add(first.get(iFirst));

        }
        for (int i = iSecond; i < second.size(); i++) {
            result.add(second.get(iSecond));

        }
        return result;
    }
}
